/*
 *  Crystality Plugin - A plugin for remastering mp3 sound in realtime
 *
 *	Copyright (C) 2002 Alex Lagutin <alex@subband.com>
 *  Copyright (C) 2001 Rafal Bosak <gyver@fanthom.irc.pl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stddef.h>
#include <math.h>
#include <string.h>
#if !defined(__osx__)
#include "mss_preflib.h"
#include "mss_dsp_plugin_api.h"
#else
#include <mss/mss_preflib.h>
#include <mss/mss_dsp_plugin_api.h>
#endif
#include "com.subband.crystality.h"
#if defined(__ppc__)
#define USE_ALTIVEC_IF_POSSIBLE
#endif

//#define	USE_HARM_BOOSTER

#define	PLUG_NAME		"Crystality"
#define	PLUG_VERSION	"1.0b1"
#define	PLUG_COPYRIGHT	"Copyright (c) 2002 Subband Software, Inc."
#define	PLUG_DESC		"Plug-in for remastering mp3 sound"
#define	PLUG_URL		"http://www.subband.com"
#if	defined(__ppc__)
	#define	PLUG_FLAGS		kMss60xRTCode | kMssPersistent
#elif defined(__i386__)
	#define	PLUG_FLAGS		kMssX86RTCode | kMssPersistent
#endif

#define	CRYSTALITY_SECTION		COM_SUBBAND_CRYSTALITY_UID
#define	BEXT_ENABLED_KEY		"ExtenderEnabled"
#define	BEXT_ENABLED_DEFAULT	1
#define	ECHO_ENABLED_KEY		"3DEchoEnabled"
#define	ECHO_ENABLED_DEFAULT	1
#define	BEXT_LEVEL_KEY			"ExtenderLevel"
#define	BEXT_LEVEL_DEFAULT		(28.0f)
#define	FILTER_LEVEL_KEY		"FilterLevel"
#define	FILTER_LEVEL_DEFAULT	3
#define	ECHO_LEVEL_KEY			"EchoLevel"
#define	ECHO_LEVEL_DEFAULT		(11.0f)
#define	STEREO_LEVEL_KEY		"StereoLevel"
#define	STEREO_LEVEL_DEFAULT	(11.0f)
#define	FEEDBACK_LEVEL_KEY		"FeedbackLevel"
#define	FEEDBACK_LEVEL_DEFAULT	(30.0f)
#define	HARMONICS_LEVEL_KEY		"HarmonicsLevel"
#define	HARMONICS_LEVEL_DEFAULT	(43.0f)

/* compile-time info */
static char	*compile_date;
static char	*compile_time;

/* Plugin API stuff */
static mssErr	Lib_Init(mssFileSpecPtr file, mssCPUType cpu);
static void		Lib_Terminate();
static mssErr	Plugin_Init(mssPluginInitParams *params, mssInstance *instance);
static void		Plugin_Terminate(mssInstance instance);
static mssErr	Plugin_Enable(mssInstance instance, mssTiming timing,
								mssHookList *hooksOut, mssHookList **hooksIn);
static void		Plugin_Disable(mssInstance instance);
static mssErr	Plugin_GetControl(mssInstance instance, uint32 controlID, void *data);
static mssErr	Plugin_SetControl(mssInstance instance, uint32 controlID, const void *data);

/* DSP Plugin API stuff */
static void		DSP_Reset(mssInstance instance, mssResetReason reason);
static void		DSP_Process(mssInstance instance, float *data, sint32 numFrames,
							mssStreamFmt *fmt, float timeStamp);

static
mssDspPluginHeader Plugin_Entry =
{
	{
		/*h.signature*/			kMssPluginSignature,
		/*h.type*/				kMssDspPluginType,
		/*h.apiVersion*/		kMssDspApiVersion,
		/*h.uid*/				COM_SUBBAND_CRYSTALITY_UID,
		/*h.name*/				PLUG_NAME,
		/*h.version*/			PLUG_VERSION,
		/*h.copyright*/			PLUG_COPYRIGHT,
		/*h.description*/		PLUG_DESC,
		/*h.url*/				PLUG_URL,
		/*h.flags*/				PLUG_FLAGS,
		/*h.reserved*/			0,
		/*h.lib_init*/			Lib_Init,
		/*h.lib_terminate*/		Lib_Terminate,
		/*h.plugin_init*/		Plugin_Init,
		/*h.plugin.term*/		Plugin_Terminate,
		/*h.plugin_enable*/		Plugin_Enable,
		/*h.plugin_disable*/	Plugin_Disable,
		/*h.get_control*/		Plugin_GetControl,
		/*h.set_control*/		Plugin_SetControl,
#if defined(__win32__)
		/*h.ui_callback*/		NULL,
#endif
	},
	/*reset*/				DSP_Reset,
	/*process*/				DSP_Process
};

MSS_EXPORT_HEADER(Plugin_Entry);

typedef struct _Interpolation
{
	int		acount;		// counter
	float	lval, rval;	// value
	float	sal, sar;	// sum
	float	al, ar;		
	float	a1l, a1r;
} Interpolation;

#if defined(USE_ALTIVEC_IF_POSSIBLE)
typedef struct _Interpolation2
{
	int				acount;	// counter
	vector float	val;	// value
	vector float	sa;		// sum
	vector float	a;		
	vector float	a1;
} Interpolation2;
#endif

class BandwidthExtender
{
	private:
	enum { kSH_BufSize = 100 * 3 };
			float		mLprev0, mRprev0, mLprev1, mRprev1,
						mLprev2, mRprev2;
			float		mLamplUp, mLamplDown, mRamplUp, mRamplDown;
		Interpolation	mEnergy,
						mAmplitude;
#if defined(USE_ALTIVEC_IF_POSSIBLE)
		Interpolation2	mAmplEnerg;
#endif
			float		bext_sfactor;
			int			filter_level;
			float		shBuf[kSH_BufSize];
			int			shBufPos, shBufPos1;
			int			cond;

	public:
						BandwidthExtender(float bextLevel, int filterLevel);
		virtual			~BandwidthExtender();
				void	Reset();
				void	Process(float *data, sint32 numFrames);
#if defined(USE_ALTIVEC_IF_POSSIBLE)
				void	ProcessVec(float *data, sint32 numFrames);
#endif
				void	ExtenderLevel(float level);
				void	FilterLevel(int level);
				
	private:
		virtual	void	pitchShifter(float lin, float rin, float *lout, float *rout);
};

#define DELAY2 21000
#define DELAY1 35000
#define DELAY3 14000
#define DELAY4 5

#define BUF_SIZE DELAY1+DELAY2+DELAY3+DELAY4

class Echo3D
{
	private:
				float	echo_sfactor,
						stereo_sfactor,
						feedback_sfactor,
						harmonics_sfactor;
				float	left0p, right0p;
				float	l0, l1, l2, r0, r1, r2;
				float	ll0, ll1, ll2, rr0, rr1, rr2;
				float	lharmb, rharmb, lhfb, rhfb;
				float	buf[BUF_SIZE];
				int		bufPos, bufPos1, bufPos2, bufPos3, bufPos4;
				
	public:
						Echo3D(float echo_level, float stereo_level,
								float feedback_level, float harmonics_level);
			virtual		~Echo3D();
				void	Reset();
				void	Process(float *data, sint32 numFrames);				
				void	EchoLevel(float echo_level);
				void	StereoLevel(float stereo_level);
				void	FeedbackLevel(float feedback_level);
				void	HarmonicsLevel(float harmonics_level);
};

class DSP_Plugin
{
	private:
			bool				mDirty;
			mssFileSpecPtr		mPrefSpecPtr;
			float				mBextLevel;
			int					mFilterLevel;
			float				mEchoLevel;
			float				mStereoLevel;
			float				mFeedbackLevel;
			float				mHarmonicsLevel;
			bool				mBextEnabled, m3DEchoEnabled;
			Echo3D				*mEcho3D;
			BandwidthExtender	*mBandExt;
		
	public:
						DSP_Plugin(mssPluginInitParams *params);
			virtual		~DSP_Plugin();
			mssErr		SetControl(uint32 controlID, const void *data);
			mssErr		GetControl(uint32 controlID, void *data);
			mssErr		Enable(mssTiming timing);
			void		Disable();
			void		Reset();
			void		Process(float *data, sint32 numFrames, mssStreamFmt *fmt);
};

#if defined(USE_ALTIVEC_IF_POSSIBLE)
int		gHasAltivec	= 0;
#endif

#if defined(USE_HARM_BOOSTER)
float	glsine[65536];
float	grsine[65536];
#endif

/* Plugin API stuff */
mssErr Lib_Init(mssFileSpecPtr file, mssCPUType cpu)
{
	UNUSED(file); UNUSED(cpu);
	compile_date	= __DATE__;
	compile_time	= __TIME__;

#if defined(USE_ALTIVEC_IF_POSSIBLE)
	gHasAltivec	= (cpu & kMssCPUAltiVecBit);
#endif

#if defined(USE_HARM_BOOSTER)	
    for (int i = 0; i < 32768; ++i)
	{
		double lsum = 0.0, rsum = 0.0;
		if (i < 32768)
			lsum	+= cos((double)i * 3.141592535 / 32768.0/2.0) / 2.0;
		if (i < 16384)
			rsum	-= cos((double)i * 3.141592535 / 16384/2.0) / 4.0;
		rsum = lsum;
	
		if (i < 8192)
			lsum	+= cos((double)i * 3.141592535 / 8192.0/2.0) / 8.0;
		if (i < 5641)
			rsum	+= cos((double)i * 3.141592535 / 5641.333333/2.0) / 8.0;
	
		glsine[32768 + i] = (double)(lsum - 0.5) * 1.45;
		glsine[32768 - i] = glsine[32768 + i];
		grsine[32768 + i] = (double)(rsum - 0.5) * 1.45;
		grsine[32768 - i] = grsine[32768 + i];
    }
#endif	
	
	return mssNoErr;
}

void Lib_Terminate()
{

}

mssErr Plugin_Init(mssPluginInitParams *params, mssInstance *instance)
{
	DSP_Plugin	*plugin;
	
	*instance	= NULL;
	if (!(plugin = new DSP_Plugin(params)))
		return mssMemFullErr;
	
	*instance	= (mssInstance) plugin;
	
	return mssNoErr;
}

void Plugin_Terminate(mssInstance instance)
{
	delete (DSP_Plugin*) instance;
}

mssErr Plugin_Enable(mssInstance instance, mssTiming timing,
					mssHookList *hooksOut, mssHookList **hooksIn)
{
	UNUSED(hooksOut); UNUSED(hooksIn);
	DSP_Plugin	*plugin = (DSP_Plugin*) instance;
	
	return plugin->Enable(timing);
}	

void Plugin_Disable(mssInstance instance)
{
	DSP_Plugin	*plugin = (DSP_Plugin*) instance;
	
	plugin->Disable();
}

mssErr Plugin_GetControl(mssInstance instance, uint32 controlID, void *data)
{
	DSP_Plugin	*plugin = (DSP_Plugin*) instance;

	return plugin->GetControl(controlID, data);
}

mssErr Plugin_SetControl(mssInstance instance, uint32 controlID, const void *data)
{
	DSP_Plugin	*plugin = (DSP_Plugin*) instance;

	return plugin->SetControl(controlID, data);
}

/* DSP Plugin API stuff */
void DSP_Reset(mssInstance instance, mssResetReason reason)
{
	DSP_Plugin	*plugin = (DSP_Plugin*) instance;
	
	plugin->Reset();
}

void DSP_Process(mssInstance instance, float *data, sint32 numFrames,
				mssStreamFmt *fmt, float timeStamp)
{
	DSP_Plugin	*plugin = (DSP_Plugin*) instance;
	
	plugin->Process(data, numFrames, fmt);
}

#if __MWERKS__
#pragma mark -
#endif

DSP_Plugin::DSP_Plugin(mssPluginInitParams *params)
:	mDirty(false),
	mPrefSpecPtr(NULL)
{
#if	defined(__macos__)
	if (params->pref)
		if ((mPrefSpecPtr = new FSSpec))
			::memcpy(mPrefSpecPtr, params->pref, sizeof(FSSpec));
#else
	if (params->pref)
		if ((mPrefSpecPtr = new char[::strlen(params->pref)+1]))
			::strcpy(mPrefSpecPtr, params->pref);
#endif
	
	mssPrefFILE	*fp = mss_pref_open(mPrefSpecPtr);
	
	mBextLevel		= mss_read_float_key(fp, CRYSTALITY_SECTION, BEXT_LEVEL_KEY, BEXT_LEVEL_DEFAULT);
	if (mBextLevel < 0.0f || mBextLevel > 100.0f)
		mBextLevel	= BEXT_LEVEL_DEFAULT;
	mFilterLevel	= mss_read_int_key(fp, CRYSTALITY_SECTION, FILTER_LEVEL_KEY, FILTER_LEVEL_DEFAULT);
	if (mFilterLevel < 1 || mFilterLevel > 3)
		mFilterLevel	= FILTER_LEVEL_DEFAULT;
	mEchoLevel		= mss_read_float_key(fp, CRYSTALITY_SECTION, ECHO_LEVEL_KEY, ECHO_LEVEL_DEFAULT);
	if (mEchoLevel < 0.0f || mEchoLevel > 100.0f)
		mEchoLevel	= ECHO_LEVEL_DEFAULT;
	mStereoLevel		= mss_read_float_key(fp, CRYSTALITY_SECTION, STEREO_LEVEL_KEY, STEREO_LEVEL_DEFAULT);
	if (mStereoLevel < 0.0f || mStereoLevel > 100.0f)
		mStereoLevel	= STEREO_LEVEL_DEFAULT;
	mFeedbackLevel		= mss_read_float_key(fp, CRYSTALITY_SECTION, FEEDBACK_LEVEL_KEY, FEEDBACK_LEVEL_DEFAULT);
	if (mFeedbackLevel < 0.0f || mFeedbackLevel > 100.0f)
		mFeedbackLevel	= FEEDBACK_LEVEL_DEFAULT;
	mHarmonicsLevel		= mss_read_float_key(fp, CRYSTALITY_SECTION, HARMONICS_LEVEL_KEY, HARMONICS_LEVEL_DEFAULT);
	if (mHarmonicsLevel < 0.0f || mHarmonicsLevel > 100.0f)
		mHarmonicsLevel	= HARMONICS_LEVEL_DEFAULT;

	mBextEnabled		= mss_read_int_key(fp, CRYSTALITY_SECTION, BEXT_ENABLED_KEY, BEXT_ENABLED_DEFAULT);
	m3DEchoEnabled		= mss_read_int_key(fp, CRYSTALITY_SECTION, ECHO_ENABLED_KEY, ECHO_ENABLED_DEFAULT);
	
	mss_pref_close(fp);
	
	mEcho3D		= NULL;
	mBandExt	= NULL;
}

DSP_Plugin::~DSP_Plugin()
{
	if (mDirty)
	{
		mssPrefFILE	*fp = mss_pref_open(mPrefSpecPtr);

		mss_save_float_key(fp, CRYSTALITY_SECTION, BEXT_LEVEL_KEY, mBextLevel);
		mss_save_int_key(fp, CRYSTALITY_SECTION, FILTER_LEVEL_KEY, mFilterLevel);
		mss_save_float_key(fp, CRYSTALITY_SECTION, ECHO_LEVEL_KEY, mEchoLevel);
		mss_save_float_key(fp, CRYSTALITY_SECTION, STEREO_LEVEL_KEY, mStereoLevel);
		mss_save_float_key(fp, CRYSTALITY_SECTION, FEEDBACK_LEVEL_KEY, mFeedbackLevel);
		mss_save_float_key(fp, CRYSTALITY_SECTION, HARMONICS_LEVEL_KEY, mHarmonicsLevel);

		mss_save_int_key(fp, CRYSTALITY_SECTION, BEXT_ENABLED_KEY, mBextEnabled);
		mss_save_int_key(fp, CRYSTALITY_SECTION, ECHO_ENABLED_KEY, m3DEchoEnabled);
		
		mss_pref_close(fp);
	}

#if	defined(__macos__)
	if (mPrefSpecPtr)
		delete mPrefSpecPtr;
#else
	if (mPrefSpecPtr)
		delete[] mPrefSpecPtr;	
#endif
}

mssErr DSP_Plugin::Enable(mssTiming timing)
{
	UNUSED(timing);
	mEcho3D		= new Echo3D(mEchoLevel, mStereoLevel, mFeedbackLevel, mHarmonicsLevel);
	mBandExt	= new BandwidthExtender(mBextLevel, mFilterLevel);
	if (!mEcho3D || !mBandExt)
	{
		if (mEcho3D)
		{
			delete mEcho3D;
			mEcho3D	= NULL;
		}
		if (mBandExt)
		{
			delete mBandExt;
			mBandExt	= NULL;
		}		
		return mssMemFullErr;
	}
	Reset();
	return mssNoErr;
}

void DSP_Plugin::Disable()
{
	if (mEcho3D)
	{
		delete mEcho3D;
		mEcho3D	= NULL;
	}
	if (mBandExt)
	{
		delete mBandExt;
		mBandExt	= NULL;
	}
}

mssErr DSP_Plugin::SetControl(uint32 controlID, const void *data)
{
	float	fval;
	int		ival;
	
	if (controlID == kCrystalitySetDefaults)
	{
		mBextLevel		= BEXT_LEVEL_DEFAULT;
		mFilterLevel	= FILTER_LEVEL_DEFAULT;
		mEchoLevel		= ECHO_LEVEL_DEFAULT;
		mStereoLevel	= STEREO_LEVEL_DEFAULT;
		mFeedbackLevel	= FEEDBACK_LEVEL_DEFAULT;
		mHarmonicsLevel	= HARMONICS_LEVEL_DEFAULT;
		
		if (!mBextEnabled)
		{
			if (mBandExt)
				mBandExt->Reset();
			mBextEnabled	= BEXT_ENABLED_DEFAULT;
		}
		if (!m3DEchoEnabled)
		{
			if (mEcho3D)
				mEcho3D->Reset();
			m3DEchoEnabled	= ECHO_ENABLED_DEFAULT;
		}
		if (mBandExt)
		{
			mBandExt->ExtenderLevel(mBextLevel);
			mBandExt->FilterLevel(mFilterLevel);
		}
		if (mEcho3D)
		{
			mEcho3D->EchoLevel(mEchoLevel);
			mEcho3D->StereoLevel(mStereoLevel);
			mEcho3D->FeedbackLevel(mFeedbackLevel);
			mEcho3D->HarmonicsLevel(mHarmonicsLevel);
		}
		mDirty	= true;
		return mssNoErr;
	}
	
	if (!data)
		return mssParamErr;
	
	switch (controlID)
	{
		case	kCrystalityBextLevel:
			fval = *(float *)data;
			
			if (fval < 0.0f || fval > 100.0f)
				return mssParamErr;

			mBextLevel	= fval;
			if (mBandExt)
				mBandExt->ExtenderLevel(mBextLevel);
			mDirty	= true;
			return mssNoErr;

		case	kCrystalityFilterLevel:
			ival = *(int *)data;
			
			if (ival < 1 || ival > 3)
				return mssParamErr;

			mFilterLevel	= ival;
			if (mBandExt)
				mBandExt->FilterLevel(mFilterLevel);
			mDirty	= true;
			return mssNoErr;

		case	kCrystalityEchoLevel:
			fval = *(float *)data;
			
			if (fval < 0.0f || fval > 100.0f)
				return mssParamErr;

			mEchoLevel	= fval;
			if (mEcho3D)
				mEcho3D->EchoLevel(mEchoLevel);
			mDirty	= true;
			return mssNoErr;

		case	kCrystalityStereoLevel:
			fval = *(float *)data;
			
			if (fval < 0.0f || fval > 100.0f)
				return mssParamErr;

			mStereoLevel	= fval;
			if (mEcho3D)
				mEcho3D->StereoLevel(mStereoLevel);
			mDirty	= true;
			return mssNoErr;

		case	kCrystalityFeedbackLevel:
			fval = *(float *)data;
			
			if (fval < 0.0f || fval > 100.0f)
				return mssParamErr;

			mFeedbackLevel	= fval;
			if (mEcho3D)
				mEcho3D->FeedbackLevel(mFeedbackLevel);
			mDirty	= true;
			return mssNoErr;

		case	kCrystalityHarmonicsLevel:
			fval = *(float *)data;
			
			if (fval < 0.0f || fval > 100.0f)
				return mssParamErr;

			mHarmonicsLevel	= fval;
			if (mEcho3D)
				mEcho3D->HarmonicsLevel(mHarmonicsLevel);
			mDirty	= true;
			return mssNoErr;

		case	kCrystalityBextEnabled:
			ival	= *(int *)data;
			if (ival && mBandExt)
				mBandExt->Reset();
			mBextEnabled = ival;
			return mssNoErr;

		case	kCrystality3DEchoEnabled:
			ival	= *(int *)data;
			if (ival && mEcho3D)
				mEcho3D->Reset();
			m3DEchoEnabled = ival;
			return mssNoErr;
	}

	return mssParamErr;
}

mssErr DSP_Plugin::GetControl(uint32 controlID, void *data)
{
	if (!data)
		return mssParamErr;
	
	switch (controlID)
	{
		case	kCrystalityBextLevel:
			*(float *)data	= mBextLevel;
			return mssNoErr;

		case	kCrystalityFilterLevel:
			*(int *)data	= mFilterLevel;
			return mssNoErr;

		case	kCrystalityEchoLevel:
			*(float *)data	= mEchoLevel;
			return mssNoErr;

		case	kCrystalityStereoLevel:
			*(float *)data	= mStereoLevel;
			return mssNoErr;

		case	kCrystalityFeedbackLevel:
			*(float *)data	= mFeedbackLevel;
			return mssNoErr;

		case	kCrystalityHarmonicsLevel:
			*(float *)data	= mHarmonicsLevel;
			return mssNoErr;

		case	kCrystalityBextEnabled:
			*(int *)data	= (int)mBextEnabled;
			return mssNoErr;

		case	kCrystality3DEchoEnabled:
			*(int *)data	= (int)m3DEchoEnabled;
			return mssNoErr;
	}
	
	return mssParamErr;
}

void DSP_Plugin::Reset()
{
	mEcho3D->Reset();
	mBandExt->Reset();
}

void DSP_Plugin::Process(float *data, sint32 numFrames, mssStreamFmt *fmt)
{
	if (!fmt->stereo || fmt->rate != 44100.0f)
		return;
	
	if (m3DEchoEnabled)
		mEcho3D->Process(data, numFrames);

	if (mBextEnabled)
	{
	#if defined(USE_ALTIVEC_IF_POSSIBLE)
		gHasAltivec ? mBandExt->ProcessVec(data, numFrames) : mBandExt->Process(data, numFrames);
	#else
		mBandExt->Process(data, numFrames);
	#endif
	}
}

#pragma mark -

Echo3D::Echo3D(float echo_level, float stereo_level, float feedback_level, float harmonics_level)
{
	EchoLevel(echo_level);
	StereoLevel(stereo_level);
	FeedbackLevel(feedback_level);
	HarmonicsLevel(harmonics_level);
}

Echo3D::~Echo3D()
{

}

void Echo3D::EchoLevel(float echo_level)
{
	echo_sfactor	= echo_level/16.0f;
}

void Echo3D::StereoLevel(float stereo_level)
{
	stereo_sfactor	= stereo_level/256.0f;
}

void Echo3D::FeedbackLevel(float feedback_level)
{
	feedback_sfactor = feedback_level/256.0f * 1.5f;
}

void Echo3D::HarmonicsLevel(float harmonics_level)
{
	harmonics_sfactor = harmonics_level;
}

void Echo3D::Reset()
{
	left0p = right0p = 0.0f;
	l0 = l1 = l2 = r0 = r1 = r2 = 0.0f;
	ll0 = ll1 = ll2 = rr0 = rr1 = rr2 = 0.0f;
	lharmb = rharmb = lhfb = rhfb = 0.0f;
	bufPos	= BUF_SIZE - 1;
	bufPos1	= 1 + BUF_SIZE - DELAY1;
	bufPos2	= 1 + BUF_SIZE - DELAY1 - DELAY2;
	bufPos3	= 1 + BUF_SIZE - DELAY1 - DELAY2 - DELAY3;
	bufPos4	= 1 + BUF_SIZE - DELAY1 - DELAY2 - DELAY3 - DELAY4;
	memset(buf, 0, sizeof(buf));
}

void Echo3D::Process(float *data, sint32 numFrames)
{
	float	left, right, dif, left0, right0, left1, right1, left2, right2,
			left3, right3, left4, right4, leftc, rightc, ls, rs;
	
	const float kf_1by8 = 1.0/8.0, kf_1by9 = 1.0/9.0, kf_1by10 = 1.0/10.0, kf_1by11 = 1.0/11.0;

#if defined(USE_HARM_BOOSTER)
	const float kf_1by32 = 1.0/32.0, kf_1by64 = 1.0/64.0, kf_1by128 = 1.0/128.0, kf_1by16384 = 1.0/16384.0;
	float	lharm0, rharm0;
	float	lsf, rsf, hsf1by64 = harmonics_sfactor * kf_1by64, hsf1by128 = harmonics_sfactor * kf_1by128;
	int		lidx, ridx;
#endif
	
	do
	{
		// ************ load sample **********
		left0	= data[0];	
		right0	= data[1];
		
		// ************ slightly expand stereo for direct input **********
		ll0	= left0;
		rr0	= right0;
		dif	= (ll0 + ll1 + ll2 - rr0 - rr1 - rr2) * stereo_sfactor;
		left0	+= dif;
		right0	-= dif;
		ll2 = ll1; ll1 = ll0;
		rr2 = rr1; rr1 = rr0;
		
		// ************ echo from buffer - first echo **********
		// ************  **********
		left1	= buf[bufPos1++];
		if (bufPos1 == BUF_SIZE)
			bufPos1 = 0;
		right1	= buf[bufPos1++];
		if (bufPos1 == BUF_SIZE)
			bufPos1 = 0;
		
		// ************ highly expand stereo for first echo **********
		dif		= (left1 - right1);
		left1	= left1 + dif;
		right1	= right1 - dif;
		
		// ************ second echo  **********
		left2	= buf[bufPos2++];
		if (bufPos2 == BUF_SIZE)
			bufPos2 = 0;	
		right2	= buf[bufPos2++];
		if (bufPos2 == BUF_SIZE)
			bufPos2 = 0;
		
		// ************ expand stereo for second echo **********
		dif		= (left2 - right2);
		left2	= left2 - dif;
		right2	= right2 - dif;
		
		// ************ third echo  **********
		left3	= buf[bufPos3++];
		if (bufPos3 == BUF_SIZE)
			bufPos3 = 0;	
		right3	= buf[bufPos3++];
		if (bufPos3 == BUF_SIZE)
			bufPos3 = 0;
		
		// ************ fourth echo  **********
		left4	= buf[bufPos4++];
		if (bufPos4 == BUF_SIZE)
			bufPos4 = 0;	
		right4	= buf[bufPos4++];
		if (bufPos4 == BUF_SIZE)
			bufPos4 = 0;
		
		left3	= (left4 + left3) * 0.5f;
		right3	= (right4 + right3) * 0.5f;
		
		// ************ expand stereo for second echo **********
		dif		= (left4 - right4);
		left3	= left3 - dif;
		right3	= right3 - dif;
		
		// ************ a weighted sum taken from reverb buffer **********
		leftc	= left1 * kf_1by9 + right2 * kf_1by8  + left3 * kf_1by8;
		rightc	= right1 * kf_1by11 + left2 * kf_1by9 + right3 *kf_1by10;
		
		// ************ mix reverb with (near to) direct input **********
		left	= left0p + leftc * echo_sfactor;
		right	= right0p + rightc * echo_sfactor;
		
		l0	= leftc + left0 * 0.5f;
		r0	= rightc + right0 * 0.5f;
		
		ls	= l0 + l1 + l2;	// do not reverb high frequencies (filter)
		rs	= r0 + r1 + r2;  //
		
		buf[bufPos++]	= ls * feedback_sfactor;
		if (bufPos == BUF_SIZE)
			bufPos = 0;
		buf[bufPos++]	= rs * feedback_sfactor;
		if (bufPos == BUF_SIZE)
			bufPos = 0;

#if defined(USE_HARM_BOOSTER)
		// ************ add some extra even harmonics **********
		// ************ or rather specific nonlinearity
		lhfb	+= (ls - lhfb) * kf_1by32;
		rhfb	+= (rs - rhfb) * kf_1by32;
		
		lsf		= ls - lhfb;
		rsf		= rs - rhfb;
#if defined(__ppc__)
		lidx	= ((((int)lsf)>>18) + 32768 + 65536) % 65536;
		ridx	= ((((int)rsf)>>18) + 32768 + 65536) % 65536;
#else
#error :FIXME:
#endif
		
		lharm0	= (lsf + MSS_POSITIVE_ONE*0.3125f) * glsine[lidx] * hsf1by64
					+ glsine[lidx] * hsf1by128;
		rharm0	= (rsf + MSS_POSITIVE_ONE*0.3125f) * glsine[ridx] * hsf1by64
					+ grsine[ridx] * hsf1by128;


		lharmb += (lharm0 - lharmb) * kf_1by16384;
		rharmb += (rharm0 - rharmb) * kf_1by16384;
				
		left  += lharm0 - lharmb;
		right += rharm0 - rharmb;
#endif

		// ************ for convolution filters **********
		l2= l1; r2= r1;
		l1 = l0; r1 = r0;
		
		left0p	= left0;
		right0p	= right0;

		// ************ store sample **********
		data[0]	= left;
		data[1]	= right;
		data	= &data[2];
		numFrames--;
	} while (numFrames);
}

#pragma mark -

BandwidthExtender::BandwidthExtender(float bextLevel, int filterLevel)
{
	ExtenderLevel(bextLevel);
	FilterLevel(filterLevel);
}

BandwidthExtender::~BandwidthExtender()
{

}

void BandwidthExtender::ExtenderLevel(float level)
{
	bext_sfactor = 32768.0f/(MSS_POSITIVE_ONE * (((16384.0f * 10.0f) / (level + 1.0f)) + (102.0f - level) * 128.0f));
}

void BandwidthExtender::FilterLevel(int level)
{
	filter_level	= level;
}

void BandwidthExtender::Reset()
{
	mLprev0 = mRprev0 = mLprev1 = mRprev1 = mLprev2 = mRprev2 = 0.0f;
	mLamplUp = mLamplDown = mRamplUp = mRamplDown = 0.0f;
	shBufPos = shBufPos1 = kSH_BufSize - 6;
	cond	= 0;
	memset(shBuf, 0, sizeof(shBuf));
	memset(&mAmplitude, 0, sizeof(mAmplitude));
	memset(&mEnergy, 0, sizeof(mEnergy));
#if defined(USE_ALTIVEC_IF_POSSIBLE)
	memset(&mAmplEnerg, 0, sizeof(mAmplEnerg));
#endif
}

/*
 * simple pitch shifter, plays short fragments at 1.5 speed
 */
 void BandwidthExtender::pitchShifter(float lin, float rin, float *lout, float *rout)
{ 
	float	*buf;
	
	shBuf[shBufPos++] = lin * 0.25f;
    shBuf[shBufPos++] = rin * 0.25f;

    if (shBufPos == kSH_BufSize)
		shBufPos = 0;
	
	buf	= &shBuf[shBufPos1];
	switch (cond)
	{
		case 1:
			*lout		= buf[0] * 2.0f + buf[2];
			*rout		= buf[1] * 2.0f + buf[3];
			break;

		case 0:
			*lout		= buf[4] * 2.0f + buf[2];
			*rout		= buf[5] * 2.0f + buf[3];
			cond		= 2;
			shBufPos1	+= 6;
			if (shBufPos1 == kSH_BufSize)
				shBufPos1 = 0;
			break;
	}
	cond--;
}

/*
 * exact bandwidth extender ("exciter") routine
 */
#if defined(USE_ALTIVEC_IF_POSSIBLE)
void BandwidthExtender::ProcessVec(float *data, sint32 numFrames)
{
	enum { kAmplCount = 64 };

	float	left, right;
    float	left0, right0, left1, right1, left2, right2, left3, right3;
	float	lprev0 = mLprev0, rprev0 = mRprev0, lprev1 = mLprev1, rprev1 = mRprev1,
			lprev2 = mLprev2, rprev2 = mRprev2;
	float	lamplUp = mLamplUp, lamplDown = mLamplDown, ramplUp = mRamplUp, ramplDown = mRamplDown;
	float	tmp;
	register vector float sval = mAmplEnerg.val, ssa = mAmplEnerg.sa;
	int		sacount = mAmplEnerg.acount;

	register const vector float	zero = (vector float)(0.0f),
								kv_1by8 = (vector float)(1.0/8.0),
								kv_120by128 = (vector float)(120.0/128.0),
								kv_AmplCount = (vector float)((float)kAmplCount),
								kv_1byAmplCount = (vector float)(1.0/(double)kAmplCount),
								kv_mss1by4 = (vector float)(MSS_POSITIVE_ONE/4.0f);
	
	register const float	kf_hpf_coeff = 56880.0/65536.0,
						kf_1000by1024 = 1000.0/1024.0,
						kf_mss1by2 = MSS_POSITIVE_ONE/2.0f,
						kf_mss4096 = MSS_POSITIVE_ONE*(4096.0/32768.0),
						kf_mss300	= MSS_POSITIVE_ONE*(300.0/32768.0),
						kf_mss500x4096 = MSS_POSITIVE_ONE*(500.0*4096.0/32768.0);
 
	union
	{
		vector float	v;
		float			f[4];
	} u;

   do
	{
		// ************ load sample **********
		left0	= data[0];
		right0	= data[1];
			
		u.f[2] = fabsf(left0  - lprev0);
		u.f[3] = fabsf(right0 - rprev0);
		// ************ highpass filter part 1 **********
		left1  = (left0  - lprev0) * kf_hpf_coeff;
		right1 = (right0 - rprev0) * kf_hpf_coeff;
		
		left2  = (left1  - lprev1) * kf_hpf_coeff;
		right2 = (right1 - rprev1) * kf_hpf_coeff;
		
		left3  = (left2  - lprev2) * kf_hpf_coeff;
		right3 = (right2 - rprev2) * kf_hpf_coeff;

		switch (filter_level)
		{
			case 1:
				pitchShifter(left1, right1, &left, &right);
				break;

			case 2:
				pitchShifter(left2, right2, &left, &right);
				break;

			case 3:
				pitchShifter(left3, right3, &left, &right);
				break;
		}

		// ************ amplitude detector ************
		tmp = left1 + lprev1;
		if      (tmp * 16.0f > lamplUp  ) lamplUp   += (tmp - lamplUp  );
		else if (tmp * 16.0f < lamplDown) lamplDown += (tmp - lamplDown);
		lamplUp		= lamplUp   * kf_1000by1024;
		lamplDown	= lamplDown * kf_1000by1024;
		u.f[0]		= fabsf(lamplUp - lamplDown);
	
		tmp = right1 + rprev1;
		if      (tmp * 16 > ramplUp  ) ramplUp   += (tmp - ramplUp  );
		else if (tmp * 16 < ramplDown) ramplDown += (tmp - ramplDown);
		ramplUp		= ramplUp   * kf_1000by1024;
		ramplDown	= ramplDown * kf_1000by1024;
		u.f[1]		= fabsf(ramplUp - ramplDown);
		
		// ************ "sound energy" detector (approx. spectrum complexity) ***********
		{
			sval	= vec_madd(kv_120by128, vec_madd(u.v, kv_1by8, sval), zero);
			ssa		= vec_add(ssa, sval);
			sacount++;
			if (sacount == kAmplCount)
			{
				vector float	a0;
				
				sacount			= 0;
				a0				= mAmplEnerg.a1;
				mAmplEnerg.a1	= vec_madd(ssa, kv_1byAmplCount, zero);
				ssa				= zero;
				mAmplEnerg.a	= vec_madd(a0, kv_AmplCount, vec_sub(mAmplEnerg.a1, a0));
			}	
			
		}

		// ************ mixer ***********
		{
			float	l, r, lsf = left * bext_sfactor, rsf = right * bext_sfactor;
			
			u.v	= vec_max(vec_min(sval, kv_mss1by4), zero);
			l	= (u.f[2] * kf_mss4096 + kf_mss500x4096) / (u.f[0] + kf_mss300) + u.f[2];
			r	= (u.f[3] * kf_mss4096 + kf_mss500x4096) / (u.f[1] + kf_mss300) + u.f[3];
			/* min(val, kf_mss1by2) */
			l	= (l - kf_mss1by2);
			l	+= fabsf(l);
			l	= kf_mss1by2 - l * 0.5f;
			r	= (r - kf_mss1by2);
			r	+= fabsf(r);
			r	= kf_mss1by2 - r * 0.5f;
					
			data[0]	= left0 + l * lsf;
			data[1]	= right0 + r * rsf;
		}

		// ************ highpass filter part 2 **********
		// ************ save previous values for filter
		lprev0 = left0;
		rprev0 = right0;
		lprev1 = left1;
		rprev1 = right1;
		lprev2 = left2;
		rprev2 = right2;
		// ************ END highpass filter part 2 **********
	
	
		data	= &data[2];
		numFrames--;
    } while (numFrames);
	
	mAmplEnerg.val		= sval;
	mAmplEnerg.sa		= ssa;
	mAmplEnerg.acount	= sacount;

	mLprev0		= lprev0;
	mRprev0		= rprev0;
	mLprev1		= lprev1;
	mRprev1		= rprev1;
	mLprev2		= lprev2;
	mRprev2		= rprev2;
	mLamplUp	= lamplUp;
	mLamplDown	= lamplDown;
	mRamplUp	= ramplUp;
	mRamplDown	= ramplDown;
}
#endif

static inline void interpolate(Interpolation *s, float l, float r)
{
	enum { kAmplCount = 64 };

	static const float	kf_1by8			= 1.0/8.0,
						kf_120by128		= 120.0/128.0,
						kf_AmplCount	= (float)kAmplCount,
						kf_1byAmplCount	= 1.0/kf_AmplCount;
	
	s->lval	+= fabsf(l) * kf_1by8;
	s->rval	+= fabsf(r) * kf_1by8;
	
	s->lval	*= kf_120by128;
	s->rval	*= kf_120by128;
	
	s->sal	+= s->lval;
	s->sar	+= s->rval;
	
	s->acount++;
	if (s->acount == kAmplCount)
	{
		float	a0l, a0r;
		
		s->acount	= 0;
		a0l			= s->a1l;
		a0r			= s->a1r;
		s->a1l		= s->sal * kf_1byAmplCount;
		s->a1r		= s->sar * kf_1byAmplCount;
		s->sal		= .0f;
		s->sar		= .0f;
		s->al		= a0l * kf_AmplCount + s->a1l - a0l;
		s->ar		= a0r * kf_AmplCount + s->a1r - a0r;
	}	
}

/*
 * calculate scalefactor for mixing and mix
 */
static inline void mix(float l0, float r0, float lin, float rin, float la,
				float ra, float le, float re, float sfactor, float *lout, float *rout)
{
	static const float	kf_mss1by4 = MSS_POSITIVE_ONE/4.0f,
						kf_mss1by2	= MSS_POSITIVE_ONE/2.0f,
						kf_mss500	= MSS_POSITIVE_ONE*(500.0/32768.0),
						kf_mss4096	= MSS_POSITIVE_ONE*(4096.0/32768.0),
						kf_mss300	= MSS_POSITIVE_ONE*(300.0/32768.0);
	
	static const float kf_2 = kf_mss500 * kf_mss4096;
	
	float	l, r, lsf = lin * sfactor, rsf = rin * sfactor;

	if (la > kf_mss1by4) la = kf_mss1by4;
	else if (la < 0.0f) la = 0.0f;
	
	if (le > kf_mss1by4) le = kf_mss1by4;
	else if (le < 0.0f) le = 0.0f;

	if (ra > kf_mss1by4) ra = kf_mss1by4;
	else if (ra < 0.0f) ra = 0.0f;
	
	if (re > kf_mss1by4) re = kf_mss1by4;
	else if (re < 0.0f) re = 0.0f;

	l	= (le * kf_mss4096 + kf_2) / (la + kf_mss300) + le;
	r	= (re * kf_mss4096 + kf_2) / (ra + kf_mss300) + re;
	
	/* min(val, kf_mss1by2) */
	l	= (l - kf_mss1by2);
	l	+= fabsf(l);
	l	= kf_mss1by2 - l * 0.5f;
	r	= (r - kf_mss1by2);
	r	+= fabsf(r);
	r	= kf_mss1by2 - r * 0.5f;
	
	*lout	= l0 + l * lsf;
	*rout	= r0 + r * rsf;
}

void BandwidthExtender::Process(float *data, sint32 numFrames)
{
	float	left, right;
    float	left0, right0, left1, right1, left2, right2, left3, right3;
	float	lprev0 = mLprev0, rprev0 = mRprev0, lprev1 = mLprev1, rprev1 = mRprev1,
			lprev2 = mLprev2, rprev2 = mRprev2;
	float	lamplUp = mLamplUp, lamplDown = mLamplDown, ramplUp = mRamplUp, ramplDown = mRamplDown;
	float	lampl, rampl;
	float	tmp;
	
	static const float	kf_hpf_coeff = 56880.0/65536.0,
						kf_1000by1024= 1000.0/1024.0;
	
    do
	{
		// ************ load sample **********
		left0	= data[0];
		right0	= data[1];
		
		// ************ highpass filter part 1 **********
		left1  = (left0  - lprev0) * kf_hpf_coeff;
		right1 = (right0 - rprev0) * kf_hpf_coeff;
		
		left2  = (left1  - lprev1) * kf_hpf_coeff;
		right2 = (right1 - rprev1) * kf_hpf_coeff;
		
		left3  = (left2  - lprev2) * kf_hpf_coeff;
		right3 = (right2 - rprev2) * kf_hpf_coeff;

		switch (filter_level)
		{
			case 1:
				pitchShifter(left1, right1, &left, &right);
				break;

			case 2:
				pitchShifter(left2, right2, &left, &right);
				break;

			case 3:
				pitchShifter(left3, right3, &left, &right);
				break;
		}

		// ************ amplitude detector ************
		tmp = left1 + lprev1;
		if      (tmp * 16.0f > lamplUp  ) lamplUp   += (tmp - lamplUp  );
		else if (tmp * 16.0f < lamplDown) lamplDown += (tmp - lamplDown);
		lamplUp   = lamplUp   * kf_1000by1024;
		lamplDown = lamplDown * kf_1000by1024;
		lampl = lamplUp - lamplDown;
	
		tmp = right1 + rprev1;
		if      (tmp * 16 > ramplUp  ) ramplUp   += (tmp - ramplUp  );
		else if (tmp * 16 < ramplDown) ramplDown += (tmp - ramplDown);
		ramplUp   = ramplUp   * kf_1000by1024;
		ramplDown = ramplDown * kf_1000by1024;
		rampl = ramplUp - ramplDown;
		
		interpolate(&mAmplitude, lampl, rampl);
	
		// ************ "sound energy" detector (approx. spectrum complexity) ***********
		interpolate(&mEnergy, left0  - lprev0, right0 - rprev0);

		// ************ mixer ***********
		mix(left0, right0, left, right, mAmplitude.lval, mAmplitude.rval,
			mEnergy.lval, mEnergy.rval, bext_sfactor, &data[0], &data[1]);
		// ************ highpass filter part 2 **********
		// ************ save previous values for filter
		lprev0 = left0;
		rprev0 = right0;
		lprev1 = left1;
		rprev1 = right1;
		lprev2 = left2;
		rprev2 = right2;
		// ************ END highpass filter part 2 **********
		
		data	= &data[2];
		numFrames--;
    } while (numFrames);

	mLprev0		= lprev0;
	mRprev0		= rprev0;
	mLprev1		= lprev1;
	mRprev1		= rprev1;
	mLprev2		= lprev2;
	mRprev2		= rprev2;
	mLamplUp	= lamplUp;
	mLamplDown	= lamplDown;
	mRamplUp	= ramplUp;
	mRamplDown	= ramplDown;
}
